__doc__ == """
**naclo** is a Python cleaning toolset for small molecule drug discovery datasets.
"""

from naclo.mol_stats import *
from naclo.mol_conversion import *
from naclo import database
from naclo import dataframes
from naclo import fragments
from naclo import neutralize
from naclo import rdpickle
from naclo import visualization
from naclo.Bleach import Bleach
from naclo.asset_loader import bleach_default_params, bleach_default_options



from rdkit import Chem
import numpy as np
from rdkit.Chem import AllChem, MACCSkeys, DataStructs


def mols_2_smiles(mols):  # *
    """Generates SMILES strings from list of rdkit Mol objects.

    Args:
        mols (iter[rdkit Mol]): Contains RDKit Mols.

    Returns:
        list[str]: Contains SMILES strings.
    """
    return [Chem.MolToSmiles(mol) for mol in mols]

def smiles_2_mols(smiles):  # *
    """Generates rdkit Mol objects from SMILES strings.

    Args:
        smiles (iter[str]): Contains SMILES strings.

    Returns:
        list[rdkit Mol]: Contains RDKit Mols.
    """
    return [Chem.MolFromSmiles(smile) for smile in smiles]

def mols_2_inchi_keys(mols):  # *
    """Generates InChI key strings from rdkit Mol objects.

    Args:
        mols (iter[rdkit Mols]): Contains rdkit Mols.

    Returns:
        list[str]: Contains InChI key strings.
    """
    return [Chem.MolToInchiKey(mol) for mol in mols]

def smiles_2_inchi_keys(smiles):  # *
    """Generates InChI key strings from SMILES strings.

    Args:
        smiles (iter[str]): Contains SMILES strings.

    Returns:
        list[str]: Contains InChI key strings.
    """
    mols = smiles_2_mols(smiles)
    return mols_2_inchi_keys(mols)

def mols_2_ecfp(mols, radius=2, return_numpy=False, n_bits=1024):
    """Converts from rdkit mol objects to morgan fingerprints (full ECFP6).

    :param mols: Collection of mols
    :type mols: list[rdkit mol]
    :param radius: ECFP radius, defaults to 3 (ECFP6)
    :type radius: int, optional
    :return: Collection of ECFP vectors
    :rtype: list[rdkit BitVect]
    """
    if return_numpy:
        fingerprints = [AllChem.GetMorganFingerprintAsBitVect(m, radius, nBits=n_bits) for m in mols]
        
        X = []
        for fp in fingerprints:
            arr = np.array([])
            DataStructs.ConvertToNumpyArray(fp, arr)
            X.append(arr)
        return X
    else:
        return [AllChem.GetMorganFingerprint(m, radius) for m in mols]

def mols_2_maccs(mols):
    """Converts from mol objects to MACCS keys.

    :param mols: Collection of molecules
    :type mols: list[rdkit mol]
    :return: Collection of MACCS keys
    :rtype: list[rdkit BitVect]
    """
    return [MACCSkeys.GenMACCSKeys(m) for m in mols]



from rdkit import Chem
from rdkit.Chem import AllChem


def init_neutralization_rxns(reactants_products={'[$([O-]);!$([O-][#7])]': 'O'}):
    """Builds reactions from dict of reactants and products.

    Args:
        reactants (dict, optional): Reactants (keys) and products (values) SMARTS subgroups. \
            Defaults to Carboxylic acids and alcohols ('[$([O-]);!$([O-][#7])]', 'O').

    Returns:
        dict: Contains Mol representations of reactants.
    """
    reactions = {}
    for reactant in reactants_products.keys():
        product = reactants_products[reactant]
        reactions[Chem.MolFromSmarts(reactant)] = Chem.MolFromSmarts(product)
        
    return reactions

def neutralize_charges(mols, reactants_products={'[$([O-]);!$([O-][#7])]': 'O'}):
    """Neutralizes Mol charges by replacing reactants with products.

    Args:
        mols (list): RDKit Mol objects. Mols to neutralize.
        reactants_products (dict, optional): Reactants (keys) and products (values) SMARTS subgroups. \
            Defaults to Carboxylic acids and alcohols ('[$([O-]);!$([O-][#7])]', 'O').

    Returns:
        list: Neutralized RDKit Mols.
    """
    reactions = init_neutralization_rxns(reactants_products=reactants_products)
    
    neutralized_mols = []
    for mol in mols:
        # Iterate over neutralization reactions
        for (reactant, product) in zip(reactions.keys(), reactions.values()):
            # Loop until all instances have been found
            while mol.HasSubstructMatch(reactant):
                mol = AllChem.ReplaceSubstructs(mol, reactant, product)[0]
                mol.UpdatePropertyCache()     
        
        neutralized_mols.append(mol)
        
    return neutralized_mols



import numpy as np
from rdkit import DataStructs


def sim_matrix(row_keys, col_keys, key_type='ecfp'):
    """Constructs a tanimoto matrix from ecfp or maccs fingerprints.

    :param row_keys: Collection of fingerprints
    :type row_keys: iter[binary]]
    :param col_keys: Collection of fingerprints 2
    :type col_keys: iter[binary]
    :param key_type: Fingerprint type, defaults to 'ecfp'
    :type key_type: str, optional
    :return: Tanimoto array
    :rtype: np.array()
    """
    
    # Initialize zero array
    full_array = np.zeros((len(row_keys), len(col_keys)))

    # Run through rows and cols of matrix according to key_type
    for i, row in enumerate(row_keys):
        for j, col in enumerate(col_keys):
            if key_type == 'ecfp':
                similarity = DataStructs.TanimotoSimilarity(row, col)
            elif key_type == 'maccs':
                similarity = DataStructs.FingerprintSimilarity(row, col)
            else:
                raise('Key type must be either "ecfp" or "maccs".')
                
            full_array[i][j] = similarity
    
    return full_array

# def plot_tsne(df, smiles_col, names_col=None, descriptors='ecfp6', alpha=1, color_col=None, continuous_grad=False):
#     seed = 42
    
#     X = []
#     error_indices = []
#     num_errors = 0
    
#     if descriptors == 'ecfp6':
#         X = ecfp6_descriptors(df[smiles_col])
        
#     elif descriptors == 'other' or descriptors == 'other_norm':
#         other_df = other_descriptors(df, [smiles_col, color_col, names_col, 'ROMol'])
#         if descriptors == 'other_norm':
#             other_df = z_norm(other_df)
#         X = other_df.to_numpy()
        
#     elif descriptors == 'ecfp6_other':
#         other_df = other_descriptors(df, [smiles_col, color_col, names_col, 'ROMol'])
#         norm_other_X = z_norm(other_df).to_numpy()
#         ecfp6_X = ecfp6_descriptors(df[smiles_col])
#         X = np.concatenate((ecfp6_X, norm_other_X), axis=1)
    
#     # Remove errors
#     fin_suppl = df.loc[~df.index.isin(error_indices)]
#     # save the full t-SNE
#     tsne = TSNE(n_components=2, random_state=seed)
#     X_fit = tsne.fit_transform(X)
#     fin_suppl['t-SNE 1'] = X_fit.T[0]
#     fin_suppl['t-SNE 2'] = X_fit.T[1]
    
#     # fin_suppl.sort_values(by=color_col, inplace=True)
    
#     if color_col:
#         fig = px.scatter(fin_suppl, x='t-SNE 1', y='t-SNE 2', color=color_col, render_mode='svg')
#     else:
#         fig = px.scatter(fin_suppl, x='t-SNE 1', y='t-SNE 2', render_mode='svg')
#     # fig = plt.figure()
#     # for cat in fin_suppl[color_col].unique():
#     #     temp = fin_suppl.where(fin_suppl[color_col] == cat)
#     #     plt.scatter(temp['tsne_x'], temp['tsne_y'], alpha=alpha)
    
#     # plt.scatter(fin_suppl['tsne_x'], fin_suppl['tsne_y'])
#     # plt.show()
#     return fig, fin_suppl


# def ecfp6_descriptors(smiles):
#     X = []
#     error_indices = []
#     num_errors = 0
    
#     for smile in smiles:
#         try:
#             mol = AllChem.MolFromSmiles(smile)
#             arr = np.zeros((0,))
#             fingerprint = AllChem.GetMorganFingerprintAsBitVect(mol, 2)
#             DataStructs.ConvertToNumpyArray(fingerprint, arr)
#             X.append(arr)
#         except:
#             num_errors += 1
#             error_indices.append(num_errors - 1)
#             pass
        
#     return X


from rdkit.Chem.Descriptors import ExactMolWt


def mol_weights(mols):  # *
    """Finds molecular weights from list of Mols.

    Args:
        mols (list): Contains RDKit Mol objects.

    Returns:
        list: Molecular weights. Numeric.
    """
    return [ExactMolWt(mol) for mol in mols]

def carbon_num(smile):  # *
    """Finds number of carbons present in SMILES string.

    Args:
        smile (str): Mol SMILES representation.

    Returns:
        int: Number of carbons in SMILES.
    """
    return smile.lower().count('c')



from argparse import ArgumentError
import pandas as pd
from stse import error_checking
from io import BytesIO, StringIO
from typing import Any, Union, Tuple
from rdkit.Chem import PandasTools
from collections.abc import Iterable


class Writer:
    def __init__(self, df:pd.DataFrame, mol_col_name:str='ROMol') -> None:
        """Initializes object for writing dataframes to various out formats.

        Args:
            df (pd.DataFrame): Data to write.
            mol_col_name (str, optional): Name of molecule column in dataframe. Defaults to 'ROMol'.
        """
        
        # Check input types
        error_checking.type_check('df', df, [pd.DataFrame])
        error_checking.type_check('mol_col_name', mol_col_name, [str])
        
        # Set properties
        self.df = df
        self.mol_col_name = mol_col_name
        
    def write(self, out:Union[str, BytesIO], ext:str) -> None:
        """Writes file to buffer object or path.

        Args:
            out (Union[str, BytesIO]): Buffer or path.
            ext (str): Save file extension.
        """
        self.__check_out(out, [str, BytesIO])
        self.__check_ext(ext)
        
        out = self.__get_writer(out, ext)
        
        if out:
            out.seek(0)
            return out
        
    def rdkit_2_excel(self, out:str, size:Tuple[int, int]=(200, 200)) -> None:
        """Writes file to Excel with MolFile images.

        Args:
            out (str): Save path.
            size (Tuple[int, int], optional): Size of image in pixels. Defaults to (200, 200).
        """
        error_checking.type_check('out', out, [str])
        error_checking.val_check('out', out.split('.')[-1], ['xlsx', 'xls'])  # Check extension is Excel
        
        PandasTools.SaveXlsxFromFrame(self.df, out, molCol=self.mol_col_name, size=size)
    
    @staticmethod
    def __check_out(out:Any, types:Iterable) -> None:
        """Wrapper for "out" type checking.

        Args:
            out (Any): Value to check type of.
            types (Iterable): Acceptable types.
        """
        error_checking.type_check('out', out, types)
    
    @staticmethod
    def __check_ext(ext:Any) -> None:
        """Wrapper for "ext" value checking.

        Args:
            ext (Any): Input to check value of.
        """
        error_checking.val_check('ext', ext, ['csv', 'tsv', 'xlsx', 'xls', 'sdf'])
    
    def __get_writer(self, out:Union[int, BytesIO], ext:str) -> Union[None, BytesIO]:
        """Writer factory.

        Args:
            out (Union[int, BytesIO]): Buffer or path.
            ext (str): Extension to determine factory piping.

        Raises:
            ArgumentError: Extension not recognized.
        """
        if ext == 'csv':
            out = self.__write_csv(out)
        elif ext == 'tsv':
            out = self.__write_tsv(out)
        elif ext == 'xlsx' or ext == 'xls':
            out = self.__write_excel(out)
        elif ext == 'sdf':
            out = self.__write_sdf(out)
        else:
            raise ArgumentError(f'ext = {self.ext} is not recognized')
        
        if out:
            return out
    
    def __write_csv(self, out:Union[int, BytesIO]) -> Union[None, BytesIO]:
        """Writes to CSV.

        Args:
            out (Union[int, BytesIO]): Buffer or path.

        Returns:
            Union[None, BytesIO]: Buffer if buffer.
        """
        self.df.to_csv(out, index=False)
        if isinstance(out, BytesIO):
            return out
        
    def __write_tsv(self, out:Union[int, BytesIO]) -> Union[None, BytesIO]:
        """Writes to TSV.

        Args:
            out (Union[int, BytesIO]): Buffer or path.

        Returns:
            Union[None, BytesIO]: Buffer if buffer.
        """
        self.df.to_csv(out, sep='\t', index=False)
        if isinstance(out, BytesIO):
            return out
        
    def __write_excel(self, out:Union[int, BytesIO]) -> Union[None, BytesIO]:
        """Writes to XLSX or XLS.

        Args:
            out (Union[int, BytesIO]): Buffer or path.

        Returns:
            Union[None, BytesIO]: Buffer if buffer.
        """
        with pd.ExcelWriter(out) as writer:
            self.df.to_excel(writer, index=False)
        if isinstance(out, BytesIO):
            return out
            
    def __write_sdf(self, out:Union[int, BytesIO]) -> Union[None, BytesIO]:
        """Writes to SDF.

        Args:
            out (Union[int, BytesIO]): Buffer or path.

        Returns:
            Union[None, BytesIO]: Buffer if buffer.
        """
        from naclo.dataframes import write_sdf  # Import within method to avoid circular import
        
        if isinstance(out, str):
            write_sdf(self.df, out, self.mol_col_name)
        else:
            out = StringIO()  # StringIO required by RDKit
            write_sdf(self.df, out, self.mol_col_name)
            out.seek(0)
            out = BytesIO(out.read().encode('utf8'))
            return out



from rdkit import Chem
from rdkit.Chem.Descriptors import ExactMolWt
from naclo import mol_stats
from rdkit.Chem.SaltRemover import SaltRemover
import numpy as np


def mw(smile):  # *
    """Removes smaller fragments from a SMILES string by molecular weight.

    Args:
        smile (str): Mol SMILES representation.

    Returns:
        str: SMILES with fragments removed.
    """
    fragments = smile.split('.')

    # Calculate molecular weights of each fragment
    mws = [ExactMolWt(Chem.MolFromSmiles(frag)) for frag in fragments]

    # Locate the max molecular weight
    max_mw = max(mws)
    max_index = mws.index(max_mw)

    return fragments[max_index]

def atom_count(smile):  # *
    """Removes smaller fragments from a SMILES string by atom count.

    Args:
        smile (str): Mol SMILES representation.

    Returns:
        str: SMILES with fragments removed.
    """
    fragments = smile.split('.')

    # Calculate atom counts of each fragment
    atom_counts = [Chem.MolFromSmiles(frag).GetNumAtoms() for frag in fragments]

    # Locate the max atom count
    max_atom_count = max(atom_counts)
    max_index = atom_counts.index(max_atom_count)

    return fragments[max_index]

def carbon_count(smile):  # *
    """Removes smaller fragments from a SMILES string by carbon count.

    Args:
        smile (str): Mol SMILES representation.

    Returns:
        str: SMILES with fragments removed.
    """
    fragments = smile.split('.')
    
    # Calculate carbon counts of each fragment
    carbon_counts = [mol_stats.carbon_num(frag) for frag in fragments]
    
    # Locate the max carbon count
    max_carbon_count = max(carbon_counts)
    max_index = carbon_counts.index(max_carbon_count)
    
    return fragments[max_index]

def remove_salts(mols, salts='[Cl,Br]'):  # *
    """Removes salts from iterable.

    Args:
        mols (iterable): Contains RDKit Mol objects.

    Returns:
        list: Contains RDKit Mol objects with salts removed.
    """
    remover = SaltRemover(defnData=salts)
    return [remover.StripMol(mol) for mol in mols]



import unittest
import json
import pandas as pd
import numpy as np
from naclo import Bleach, bleach_default_options, bleach_default_params
import warnings
from rdkit import Chem
import copy


class TestBleach(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.default_params = bleach_default_params
        cls.default_options = bleach_default_options
            
        cls.smiles_df = pd.DataFrame({
            'SMILES': ['Cc1cc(/C=C/C#N)cc(C)c1Nc1nc(Nc2ccc(C#N)cc2)ncc1N',
                       'Cc1cc(/C=C/C#N)cc(C)c1Nc1ncc(N)c(Nc2c(C)cc(/C=C/C#N)cc2C)n1',
                       'CCC',
                       'CCC.Cl',
                       'C.NO.S',
                       'Br',
                       '']
        })
        
        warnings.filterwarnings('error')  # Catch in try except
        
        return super().setUpClass()
    
    def test_param_checker(self):
        params = copy.deepcopy(self.default_params)
        options = copy.deepcopy(self.default_options)
        
        try:
            Bleach(self.smiles_df, params, options)
        except ValueError as e:
            self.assertEqual(
                e.args[0],
                'NO_STRUCTURE_COLUMN'
            )
        
        params['structure_col'] = 'test'
        try:
            Bleach(self.smiles_df, params, options)
        except ValueError as e:
            self.assertEqual(
                e.args[0],
                'STRUCTURE_COLUMN_NOT_FOUND'
            )
        
        params['structure_col'] = 'SMILES'
        try:
            Bleach(self.smiles_df, params, options)
        except ValueError as e:
            self.assertEqual(
                e.args[0],
                'INVALID_STRUCTURE_TYPE'
            )
        
        params['structure_type'] = 'smiles'
        params['target_col'] = 'test'
        try:
            Bleach(self.smiles_df, params, options)
        except ValueError as e:
            self.assertEqual(
                e.args[0],
                'TARGET_COLUMN_NOT_FOUND'
            )
        
        # Should run fine
        params['target_col'] = 'SMILES'
        Bleach(self.smiles_df, params, options)
        
    def test_recognized_options_checker(self):
        params = copy.deepcopy(self.default_params)
        options = copy.deepcopy(self.default_options)
        params['structure_col'] = 'SMILES'
        params['structure_type'] = 'smiles'
        options['molecule_settings']['neutralize_charges']['run'] = 42
        
        try:
            Bleach(self.smiles_df, params, options)
        except ValueError as e:
            self.assertEqual(
                list(e.args[0].keys())[0],
                'BAD_OPTION_MOLECULE_SETTINGS_NEUTRALIZE_CHARGES_RUN'
            )

    def test_drop_na(self):
        df = pd.DataFrame({
            'SMILES': [
                pd.NA,
                '',
                'nan',
                'none',
                'CCC',
                np.nan
            ],
            'ROMol': [
                pd.NA,
                '',
                'nan',
                'C',
                'none',
                np.nan
            ],
            'drop_empty': 6*[np.nan]
        })
        
        params = copy.deepcopy(self.default_params)
        
        # SMILES
        params['structure_col'] = 'SMILES'
        params['structure_type'] = 'smiles'
        
        bleach = Bleach(df, params, self.default_options)
        bleach.drop_na()
        
        expected = pd.DataFrame({
            'SMILES': ['CCC']
        }, index=[4])
        
        self.assertEqual(
            True,
            bleach.df.equals(expected)
        )
        
        # ROMol
        params['structure_col'] = 'ROMol'
        params['structure_type'] = 'mol'
        
        bleach = Bleach(df, params, self.default_options)
        bleach.drop_na()
        
        expected = pd.DataFrame({
            'ROMol': ['C']
        }, index=[3])
        
        self.assertEqual(
            True,
            bleach.df.equals(expected)
        )
        
        # ALL_NA_STRUCTURES warning
        params['structure_col'] = 'drop_empty'
        
        bleach = Bleach(df, params, self.default_options)
        
        try:
            bleach.drop_na()
        except RuntimeWarning as w:
            self.assertEqual(
                w.args[0],
                'ALL_NA_STRUCTURES: All structures in specified column were NA, all rows dropped'
            )
            
        # ALL_NA_TARGETS warning
        params['structure_col'] = 'ROMol'
        params['target_col'] = 'drop_empty'
        
        # Not set to drop NA targets
        bleach = Bleach(df, params, self.default_options)
        bleach.drop_na()
        
        expected = pd.DataFrame({
            'ROMol': ['C']
        }, index=[3])
        
        self.assertEqual(
            True,
            bleach.df.equals(expected)
        )
        
        # Set to drop NA targets
        options = copy.deepcopy(self.default_options)
        options['file_settings']['remove_na_targets']['run'] = True
        
        bleach = Bleach(df, params, options)
        
        try:
            bleach.drop_na()
        except RuntimeWarning as w:
            self.assertEqual(
                w.args[0],
                'ALL_NA_TARGETS: All targets in specified column were NA, all rows dropped'
            )
            
    def test_init_structure_compute(self):
        # From SMILES
        params = copy.deepcopy(self.default_params)
        params['structure_col'] = 'SMILES'
        params['structure_type'] = 'smiles'
        
        bleach = Bleach(self.smiles_df, params, self.default_options)
        bleach.drop_na()
        bleach.init_structure_compute()

        [self.assertIsInstance(m, Chem.rdchem.Mol) for m in bleach.df['ROMol']]
        
        # From SMILES w/ bad Mols
        mol_df = bleach.df.copy()
        
        new_row = pd.DataFrame({
            'SMILES': ['CC', 'C', 'CCC'],
            'ROMol': ['test', 1, 'CCC']
        })
        
        mol_df = pd.concat((mol_df, new_row))
        bleach = Bleach(mol_df, params, self.default_options)
        bleach.drop_na()
        bleach.init_structure_compute()
        
        self.assertEqual(
            bleach.df['SMILES'].tolist(),
            mol_df['SMILES'].tolist()
        )
        [self.assertIsInstance(m, Chem.rdchem.Mol) for m in bleach.df['ROMol']]
        
        # From Mols w/ bad Mols
        params['structure_col'] = 'ROMol'
        params['structure_type'] = 'mol'
        bleach = Bleach(mol_df, params, self.default_options)
        bleach.drop_na()
        bleach.init_structure_compute()
        
        [self.assertIsInstance(m, Chem.rdchem.Mol) for m in bleach.df['ROMol']]
        self.assertEqual(
            len(bleach.df),
            6
        )
        self.assertEqual(
            bleach.df['SMILES'].tolist(),
            self.smiles_df['SMILES'].tolist()[:-1]  # Excluding blank
        )
        
    def test_mol_cleanup(self):
        params = copy.deepcopy(self.default_params)
        options = copy.deepcopy(self.default_options)
        params['structure_col'] = 'SMILES'
        params['structure_type'] = 'smiles'
        
        # Salts only
        options['molecule_settings']['remove_fragments']['filter_method'] = 'none'
        
        bleach = Bleach(self.smiles_df, params, options)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.mol_cleanup()
        
        expected = pd.DataFrame({
            'SMILES': ['Cc1cc(/C=C/C#N)cc(C)c1Nc1nc(Nc2ccc(C#N)cc2)ncc1N',
                       'Cc1cc(/C=C/C#N)cc(C)c1Nc1ncc(N)c(Nc2c(C)cc(/C=C/C#N)cc2C)n1',
                       'CCC',
                       'CCC',
                       'C.NO.S']
        })
        
        self.assertTrue(
            bleach.df['SMILES'].equals(expected['SMILES'])
        )
        
        # Filters only
        options['molecule_settings']['remove_fragments']['salts'] = ''
        
        # Carbon count
        options['molecule_settings']['remove_fragments']['filter_method'] = 'carbon_count'
        bleach = Bleach(self.smiles_df, params, options)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.mol_cleanup()
        self.assertEqual(
            bleach.df['SMILES'].iloc[-2],
            'C'
        )
        
        # Atom count
        options['molecule_settings']['remove_fragments']['filter_method'] = 'atom_count'
        bleach = Bleach(self.smiles_df, params, options)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.mol_cleanup()
        self.assertEqual(
            bleach.df['SMILES'].iloc[-2],
            'NO'
        )
        
        # Molecular weight
        options['molecule_settings']['remove_fragments']['filter_method'] = 'mw'
        bleach = Bleach(self.smiles_df, params, options)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.mol_cleanup()
        self.assertEqual(
            bleach.df['SMILES'].iloc[-2],
            'S'
        )
        
        # Salts + filter together
        options['molecule_settings']['remove_fragments']['salts'] = 'Cl,Br'
        options['molecule_settings']['remove_fragments']['filter_method'] = 'carbon_count'
        
        bleach = Bleach(self.smiles_df, params, options)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.mol_cleanup()
        
        expected = pd.DataFrame({
            'SMILES': ['Cc1cc(/C=C/C#N)cc(C)c1Nc1nc(Nc2ccc(C#N)cc2)ncc1N',
                       'Cc1cc(/C=C/C#N)cc(C)c1Nc1ncc(N)c(Nc2c(C)cc(/C=C/C#N)cc2C)n1',
                       'CCC',
                       'CCC',
                       'C']
        })
        
        self.assertTrue(
            bleach.df['SMILES'].equals(expected['SMILES'])
        )
        
    def test_handle_duplicates(self):
        params = copy.deepcopy(self.default_params)
        options = copy.deepcopy(self.default_options)
        params['structure_col'] = 'SMILES'
        params['structure_type'] = 'smiles'
        params['target_col'] = 'target'
        
        df = self.smiles_df.copy()
        df['target'] = [1, 2, 3, 4, 5, 6, 7]
        
        # Average
        bleach = Bleach(df, params, options)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.mol_cleanup()
        bleach.handle_duplicates()
        
        expected = pd.DataFrame({
            'SMILES': ['Cc1cc(/C=C/C#N)cc(C)c1Nc1nc(Nc2ccc(C#N)cc2)ncc1N',
                       'Cc1cc(/C=C/C#N)cc(C)c1Nc1ncc(N)c(Nc2c(C)cc(/C=C/C#N)cc2C)n1',
                       'CCC',
                       'C'],
            
            'InchiKey': [
                'NFQWIYBXPYHNRC-ONEGZZNKSA-N',
                'IKTVSGAURLVDFP-KQQUZDAGSA-N',
                'ATUOYWHBWRKTHZ-UHFFFAOYSA-N',
                'VNWKTOKETHGBQD-UHFFFAOYSA-N'],
            
            'target': [1,
                       2,
                       3.5,
                       5]
        })
        
        self.assertTrue(
            bleach.df[['SMILES', 'InchiKey', 'target']].reset_index(drop=True).equals(expected)
        )
        
        # Remove
        options['file_settings']['duplicate_compounds']['selected'] = 'remove'
        bleach = Bleach(df, params, options)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.mol_cleanup()
        bleach.handle_duplicates()
        
        expected['target'] = [1, 2, 3, 5]
        
        self.assertTrue(
            bleach.df[['SMILES', 'InchiKey', 'target']].reset_index(drop=True).equals(expected)
        )
        
        # Keep
        options['file_settings']['duplicate_compounds']['selected'] = 'keep'
        bleach = Bleach(df, params, options)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.mol_cleanup()
        
        df1 = bleach.df.copy()
        bleach.handle_duplicates()
        df2 = bleach.df.copy().drop('InchiKey', axis=1)
        
        self.assertTrue(
            df1.equals(df2)
        )
        
        self.assertIn(
            'InchiKey',
            bleach.df.columns
        )
        
    def test_append_column(self):
        params = copy.deepcopy(self.default_params)
        options = copy.deepcopy(self.default_options)
        params['structure_col'] = 'SMILES'
        params['structure_type'] = 'smiles'
        
        # All columns
        bleach = Bleach(self.smiles_df, params, options)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.mol_cleanup()
        bleach.handle_duplicates()
        bleach.append_columns()
        
        self.assertEqual(
            ['SMILES', 'ROMol', 'InchiKey', 'MW'],
            list(bleach.df.columns)
        )
        
        # No columns (except SMILES for testing purposes)
        options['file_settings']['append_columns'] = {
            'smiles': True,
            'mol': False,
            'inchi_key': False,
            'mw': False
        }
        
        
        bleach = Bleach(self.smiles_df, params, options)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.mol_cleanup()
        bleach.handle_duplicates()
        bleach.append_columns()
        
        self.assertEqual(
            ['SMILES'],
            list(bleach.df.columns)
        )
        
    def test_remove_header_chars(self):
        params = copy.deepcopy(self.default_params)
        options = copy.deepcopy(self.default_options)
        params['structure_col'] = 'SMILES'
        params['structure_type'] = 'smiles'
        
        options['file_settings']['remove_header_chars']['chars'] = 'sr'
        
        bleach = Bleach(self.smiles_df, params, options)
        bleach.main()
        
        self.assertEqual(
            ['MILE', 'OMol', 'InchiKey', 'MW'],
            list(bleach.df.columns)
        )


if __name__ == '__main__':
    unittest.main()



import unittest
from rdkit import Chem
from rdkit.Chem import MACCSkeys
from naclo import mol_conversion, visualization
import numpy as np


class TestVisualization(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.test_smiles = ['CN=C=O', 'CCC', 'O']
        cls.test_mols = mol_conversion.smiles_2_mols(cls.test_smiles)
        cls.maccs_keys = mol_conversion.mols_2_maccs(cls.test_mols)
        cls.ecfp4_prints = mol_conversion.mols_2_ecfp(cls.test_mols, radius=2, return_numpy=False)
        cls.ecfp6_prints = mol_conversion.mols_2_ecfp(cls.test_mols, radius=3, return_numpy=False)
        return super().setUpClass()
    
    def test_sim_matrix(self):
        mat_maccs = visualization.sim_matrix(self.maccs_keys, self.maccs_keys, key_type='maccs')  # Test MACCS keys
        mat_ecfp4 = visualization.sim_matrix(self.ecfp4_prints, self.ecfp4_prints)  # Test ECFP4
        mat_ecfp6 = visualization.sim_matrix(self.ecfp6_prints, self.ecfp6_prints)  # Test ECFP6
        
        # Check same molecules equal 1 in matrix in all cases (along identity)
        np.testing.assert_array_equal(
            mat_maccs[np.eye(3, dtype=bool)],
            mat_ecfp4[np.eye(3, dtype=bool)],
            mat_ecfp6[np.eye(3, dtype=bool)],
            3*[1]
        )
        
        # Check maccs is different than ecfp but ecfp4,6 are equal
        self.assertFalse(
            np.array_equal(
                mat_maccs,
                mat_ecfp4
            )
        )
        np.testing.assert_array_equal(
            mat_ecfp4,
            mat_ecfp6
        )
        
        
if __name__ == '__main__':
    unittest.main()



import json
import pandas as pd
import warnings
from typing import Callable

# sourced from github.com/jwgerlach00
import naclo
import stse


class Bleach:
    def __init__(self, df:pd.DataFrame, params:dict, options:dict) -> None:  # *
        # Load user options
        self.mol_settings = options['molecule_settings']
        self.file_settings = options['file_settings']
        self.__recognized_options_checker()

        self.__recognized_structures = ['smiles', 'mol']
        self.__default_cols = {
            'smiles': 'SMILES',
            'mol': 'ROMol',
            'inchi_key': 'InchiKey',
            'mw': 'MW'
        }

        # Save user input data
        self.original_df = df.copy()
        self.df = df.copy()

        # Load file parameters
        self.structure_col = params['structure_col']
        self.structure_type = params['structure_type']
        self.target_col = params['target_col']
        self.__param_checker()

        self.mol_col = None
        self.smiles_col = None
        self.__set_structure_cols()  # Assign mol and SMILES cols using input + defaults
        self.inchi_key_col = None


# -------------------------------------------------- ERROR CHECKING -------------------------------------------------- #
    def __recognized_options_checker(self) -> None:  # *
        """Checks for errors related to unrecognized option assignments.

        Raises:
            ValueError: BAD_OPTION(S)
        """
        with open('naclo/assets/recognized_bleach_options.json') as f:
            recognized_options = json.load(f)

        input = stse.dictionaries.branches({
            'molecule_settings': self.mol_settings,
            'file_settings': self.file_settings
        })
        recognized = stse.dictionaries.branches(recognized_options)

        errors = {}
        for key, value in recognized.items():
            if isinstance(value, list):
                if not input[key] in recognized[key]:
                    errors[f'BAD_OPTION{key.upper()}'] = f'"{input[key]}" is not an accepted value for "{key}", set \
                        to one of: "{recognized[key]}"'
            else:
                if not type(input[key]) == type(recognized[key]):
                    errors[f'BAD_OPTION{key.upper()}'] = f'{type(input[key])} is not an accepted type for {key}, \
                        input a {type(recognized[key])}'
        if errors:
            raise ValueError(errors)

    def __param_checker(self) -> None:  # *
        """Checks for errors related to declared parameters.

        Raises:
            ValueError: NO_STRUCTURE_COLUMN
            ValueError: STRUCTURE_COLUMN_NOT_FOUND
            ValueError: INVALID_STRUCTURE_TYPE
            ValueError: TARGET_COLUMN_NOT_FOUND
        """
        if not self.structure_col:
            raise ValueError('NO_STRUCTURE_COLUMN', 'Must specify the name of the structure column in params')

        if self.structure_col not in self.df.columns:
            raise ValueError('STRUCTURE_COLUMN_NOT_FOUND', f'The structure column: "{self.structure_col}"" is not \
                present in the data: "{list(self.df.columns)}"')

        if self.structure_type not in self.__recognized_structures:
            raise ValueError('INVALID_STRUCTURE_TYPE', f'Structure type: "{self.structure_type}"" is not one of: \
                {self.__recognized_structures}')

        if self.target_col and self.target_col not in self.df.columns:
            raise ValueError('TARGET_COLUMN_NOT_FOUND', f'The target column: "{self.target_col}"" is not present in \
                the data: "{list(self.df.columns)}"')


# -------------------------------------------------- STATIC METHODS -------------------------------------------------- #
    @staticmethod
    def __filter_fragments_factory(filter:str) -> Callable:
        """Returns a callable SMILES fragment filter function using a key.

        Args:
            filter (str): Function key.

        Raises:
            ValueError

        Returns:
            Callable: Filter function.
        """
        if filter == 'carbon_count':
            return naclo.fragments.carbon_count
        elif filter == 'mw':
            return naclo.fragments.mw
        elif filter == 'atom_count':
            return naclo.fragments.atom_count
        else:
            raise ValueError('Filter method is not allowed')


# -------------------------------------------------- PRIVATE METHODS ------------------------------------------------- #
    def __set_structure_cols(self) -> None:
        """Sets Mol and SMILES columns using declared structure type."""
        self.mol_col = self.structure_col if self.structure_type == 'mol' else self.__default_cols['mol']
        self.smiles_col = self.structure_col if self.structure_type == 'smiles' else self.__default_cols['smiles']

    def __drop_na_structures(self) -> None:
        """Drops NA along declared structure column."""
        self.df.dropna(subset=[self.structure_col], inplace=True)
        if not len(self.df):
            warnings.warn('ALL_NA_STRUCTURES: All structures in specified column were NA, all rows dropped',
                          RuntimeWarning)

    def __drop_na_targets(self) -> None:
        """Drops NA along declared target column"""
        run_na_targets = self.file_settings['remove_na_targets']['run']

        if self.target_col and run_na_targets and len(self.df):  # If run and TARGET COLUMN DECLARED
            self.df.dropna(subset=[self.target_col], inplace=True)
            if not len(self.df):
                warnings.warn('ALL_NA_TARGETS: All targets in specified column were NA, all rows dropped',
                              RuntimeWarning)

        elif run_na_targets:  # If run but not declared target
            warnings.warn('NA_TARGETS: options.file_settings.remove_na_targets was set to run but no activity column \
                was specified', RuntimeWarning)

    def __build_smiles(self) -> None:
        """Creates a SMILES column in the dataset using dataset MolFile column. DROPS NA."""
        self.df = naclo.dataframes.df_mols_2_smiles(self.df, self.mol_col, self.smiles_col)

    def __build_mols(self) -> None:
        """Creates MolFile column in the dataset using dataset SMILES column. DROPS NA."""
        self.df = naclo.dataframes.df_smiles_2_mols(self.df, self.smiles_col, self.mol_col)

    def __remove_fragments(self) -> None:
        """Removes salts if specified. Drops NA as a result of salt removal. Filters out other fragments by specified
        method."""
        option = self.mol_settings['remove_fragments']

        # Remove salts
        if option['salts']:
            self.df[self.mol_col] = naclo.fragments.remove_salts(self.df[self.mol_col], salts='[{0}]'.format(
                option['salts'].replace(' ', '')))
            self.__build_smiles()

            # Drop NA (blank string after salts)
            self.df = stse.dataframes.convert_to_nan(self.df, na=[''])  # Convert bc NA is just empty string
            self.df.dropna(subset=[self.smiles_col], inplace=True)  # Drop NA bc may include molecule that is ONLY salts

        # Filter
        if option['filter_method'] and option['filter_method'] != 'none':
            self.df[self.smiles_col] = self.df[self.smiles_col].apply(
                self.__filter_fragments_factory(option['filter_method']))
            self.__build_mols()

    def __neutralize_charges(self) -> None:
        """Neutralizes Mols. Rebuilds SMILES."""
        self.df[self.mol_col] = naclo.neutralize.neutralize_charges(self.df[self.mol_col])
        self.__build_smiles

    def __compute_inchi_keys(self) -> None:
        """Declares inchi key column name using default. Appends inchi keys to dataset."""
        self.inchi_key_col = self.__default_cols['inchi_key']
        self.df = naclo.dataframes.df_mols_2_inchi_keys(self.df, self.mol_col, self.inchi_key_col)

    def __drop_columns(self) -> None:
        """Removes columns that the user does not want in the final output."""
        option = self.file_settings['append_columns']

        # Drop added columns from built if not requested
        if not option['mol']:
            self.df.drop(self.mol_col, inplace=True, axis=1)
        if not option['inchi_key']:
            self.df.drop(self.inchi_key_col, inplace=True, axis=1)
        if not option['smiles']:
            self.df.drop(self.smiles_col, inplace=True, axis=1)

    def __add_columns(self) -> None:
        """Add columns that the user wants in the final output."""
        option = self.file_settings['append_columns']
        
        # Add MW column
        if option['mw']:
            self.df = self.df.assign(MW = naclo.mol_weights(self.df[self.mol_col]))


# ------------------------------------------------- PUBLIC FUNCTIONS ------------------------------------------------- #
    # Step 1
    def drop_na(self) -> None:  # *
        """Converts blanks to NA. Drops NA Mols or SMILES. Handles NA targets. Removes entire NA columns"""

        # Convert all df blanks and 'none' to NA
        self.df = stse.dataframes.convert_to_nan(self.df)

        # Drop rows
        self.__drop_na_structures()
        self.__drop_na_targets()

        # Drop cols
        self.df = stse.dataframes.remove_nan_cols(self.df)  # After dropping rows because columns may BECOME empty

    # Step 2 
    def init_structure_compute(self) -> None:  # *
        """Builds (or rebuilds from Mols) SMILES. Builds Mols if not present in dataset."""
        if self.structure_type == 'mol':
            self.__build_smiles()
            # Rebuilding Mols not necessary

        elif self.structure_type == 'smiles':
            self.__build_mols()
            self.__build_smiles()  # Canonicalize SMILES

    # Step 3
    def mol_cleanup(self):  # *
        """Cleans Mols and SMILES."""

        # Step 1: Deal with fragments (includes salt step -- may include a molecule that is ONLY salts (NA dropped))
        self.__remove_fragments()

        # Step 2: Neutralize mols
        if self.mol_settings['neutralize_charges']['run']:
            self.__neutralize_charges()

    # Step 4
    def handle_duplicates(self):  # *
        """Computes inchi keys. Averages, removes, or keeps duplicates. ONLY BY INCHI KEY FOR NOW."""
        self.__compute_inchi_keys()

        dup = self.file_settings['duplicate_compounds']

        if dup['selected'] == 'average' and self.target_col:
            self.df = stse.duplicates.average(self.df, subsets=[self.inchi_key_col], average_by=self.target_col)
        elif dup['selected'] == 'remove' or (dup['selected'] == 'average' and not self.target_col):
            self.df = stse.duplicates.remove(self.df, subsets=[self.inchi_key_col])

    # Step 5
    def append_columns(self) -> None:  # *
        """Drops and adds columns depending on what the user wants returned.

        Args:
            df (pandas DataFrame): Data to transform
        """
        self.__drop_columns()
        self.__add_columns()

    # Step 6
    def remove_header_chars(self) -> None:  # *
        """Removes any chars listed in a string of chars from the df column headers.
        """
        chars = self.file_settings['remove_header_chars']['chars']
        self.df = stse.dataframes.remove_header_chars(self.df, chars)


# ----------------------------------------------------- MAIN LOOP ---------------------------------------------------- #
    def main(self) -> pd.DataFrame:
        """Main bleach loop.

        Returns:
            pandas DataFrame: Cleaned df
        """
        self.drop_na()  # Before init_structure bc need NA
        self.init_structure_compute()
        self.mol_cleanup()
        self.handle_duplicates()  # Includes InchiKey generation
        self.append_columns()
        self.remove_header_chars()
        
        return self.df



from rdkit import Chem
from warnings import warn
from rdkit.Chem import PandasTools
import pandas as pd
from typing import Union
import numpy as np

# Nested imports
from naclo.Writer import Writer


def __exception_2_nan(x, func):
        try:
            return func(x)
        except Exception:
            return np.nan
    

def df_smiles_2_mols(df, smiles_name, mol_name, dropna=True):  # *
    """Adds rdkit Mol column to df using SMILES column as reference.

    Args:
        df (pandas DataFrame): DataFrame to add Mol column to.
        smiles_name (str): Name of SMILES column in df.
        molecule_name (str): Name of Mol column in df.
        dropna (bool, optional): Drop NA Mols. Defaults to True.

    Returns:
        pandas DataFrame: DataFrame with Mol column appended.
    """
    df[mol_name] = df[smiles_name].map(lambda x: __exception_2_nan(x, Chem.MolFromSmiles), na_action='ignore')
    return df.dropna(subset=[mol_name]) if dropna else df

def df_mols_2_inchi_keys(df, mol_name, inchi_name, dropna=True):  # *
    """Adds InChi Key column to df using Mol column as reference.
    
    Args:
        df (pandas DataFrame): DataFrame to add InChi column to.
        mol_name (str): Name of InChi column in df.
        inchi_name (str): Name of InChi column in df.
        dropna (bool, optional): Drop NA InChis. Defaults to True.

    Returns:
        pandas DataFrame: DataFrame with InChi column appended.
    """
    df[inchi_name] = df[mol_name].map(lambda x: __exception_2_nan(x, Chem.MolToInchiKey), na_action='ignore')
    return df.dropna(subset=[inchi_name]) if dropna else df
    
def df_mols_2_smiles(df, mol_name, smiles_name, dropna=True):  # *
    """Adds SMILES Key column to df using Mol column as reference.

    Args:
        df (pandas DataFrame): DataFrame to add SMILES column to.
        mol_name (str): Name of Mol column in df.
        inchi_name (str): Name of InChi column in df.
        dropna (bool, optional): Drop NA SMILES. Defaults to True.

    Returns:
        pandas DataFrame: DataFrame with SMILES column appended.
    """
    df[smiles_name] = df[mol_name].map(lambda x: __exception_2_nan(x, Chem.MolToSmiles), na_action='ignore')
    return df.dropna(subset=[smiles_name]) if dropna else df

def write_sdf(df, out_path, mol_col_name, id_column_name='RowID'):  # *
    """Writes dataframe to SDF file. Includes ID name if ID is valid.

    Args:
        df (DataFrame): Data to write.
        out_path (str or file-like): Path to save SDF to.
        mol_col_name (str): Name of molecule column in dataframe.
        id_column_name (str, optional): Name of id column. Defaults to 'ID'.
    """
    try:
        PandasTools.WriteSDF(df, out_path, molColName=mol_col_name, properties=df.columns, idName=id_column_name)
    except KeyError:
        PandasTools.WriteSDF(df, out_path, molColName=mol_col_name, properties=df.columns, idName='RowID')
        warn(f'write_sdf \'{id_column_name}\' ID name invalid', UserWarning)

def id_mol_col(df:pd.DataFrame) -> Union[None, pd.Index]:
    """Identifies first column that contains a Mol object based on first row.

    Args:
        df (pd.DataFrame): Input data.

    Returns:
        Union[None, pd.Index]: Index of first Mol column if found. Else None.
    """
    for i in range(len(df)):
        df.iloc[i]
        first_row = [isinstance(m, Chem.rdchem.Mol) for m in df.iloc[i]]
        if sum(first_row) > 0:
            return df.columns[first_row.index(True)]
        
    return None



from importlib import resources
import json


with resources.open_text('naclo.assets', 'default_params.json') as f:
    bleach_default_params = json.load(f)

with resources.open_text('naclo.assets', 'default_options.json') as f:
    bleach_default_options = json.load(f)
